# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import datetime
import re
import sys

from . import util


if sys.version_info >= (3, 7):

    def _parse_iso_date(date_string):
        return datetime.date.fromisoformat(date_string)

else:

    def _parse_iso_date(date_string):
        m = re.match("([0-9]{4})-([0-9]{2})-([0-9]{2})", date_string)
        assert m
        return datetime.date(int(m.group(1)), int(m.group(2)), int(m.group(3)))


def _parse_date(date_string):
    if not date_string:
        return None
    return _parse_iso_date(date_string)


@util.os_parametrize("osxml", filter_dates=True)
def test_dates(osxml):
    release_date = _parse_date(osxml.release_date)
    eol_date = _parse_date(osxml.eol_date)
    if release_date and eol_date:
        assert release_date < eol_date



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import util


@util.os_parametrize("osxml", filter_media=True)
def test_existing_variant_for_media(osxml):
    variants = osxml.variants
    for media in osxml.medias:
        variant = media.variant
        if variant and variant not in variants:
            raise AssertionError(
                "variants: media in OS %s has variant=%s "
                "but declared variants are %s" % (osxml.shortid, variant, variants)
            )


@util.os_parametrize("osxml", filter_trees=True)
def test_existing_variant_for_trees(osxml):
    variants = osxml.variants
    for tree in osxml.trees:
        variant = tree.variant
        if variant and variant not in variants:
            raise AssertionError(
                "variants: tree in OS %s has variant=%s "
                "but declared variants are %s" % (osxml.shortid, variant, variants)
            )


@util.os_parametrize("osxml", filter_images=True)
def test_existing_variant_for_images(osxml):
    variants = osxml.variants
    for image in osxml.images:
        variant = image.variant
        if variant and variant not in variants:
            raise AssertionError(
                "variants: image in OS %s has variant=%s "
                "but declared variants are %s" % (osxml.shortid, variant, variants)
            )



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import util


@util.os_parametrize("osxml", filter_related=True)
def test_related(osxml):
    related = util.DataFiles.getosxml_related(osxml)
    upgrades = osxml.upgrades
    if upgrades:
        found = util.DataFiles.getosxml_by_id(upgrades)
        assert found
        related.append(found)

    for os in related:
        assert os is not osxml



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import re

from . import util


DEVICE_MAP = {d.internal_id: d for d in util.DataFiles.devices()}
DEVICE_MAP_SRC = {d.internal_id: d for d in util.SourceFiles.devices()}


def _check_duplicate_devices(osxml):
    """
    Ensure an OS doesn't list a device that's enabled, and not eventually
    disabled, in a parent OS
    """
    related = util.DataFiles.getosxml_related(osxml)
    inherited_devices = set()
    for osxml2 in reversed(related):
        inherited_devices.update(set(osxml2.devices))
        inherited_devices.difference_update(osxml2.devices_unsupported)

    assert set(osxml.devices).intersection(inherited_devices) == set()


def _check_uncommented_devices(osxml):
    """
    Ensure every device listed in the XML is followed by a comment with
    the device string name in it. This helps readability/grepability
    """
    badlines = []
    devlines = [
        line
        for line in osxml.path.open("r").read().splitlines()
        if "<device id" in line
    ]

    for devid in osxml.devices:
        devname = DEVICE_MAP_SRC[devid].name
        for devline in devlines:
            if devid not in devline:
                continue
            if not re.search(r"<!--.*%s.*-->" % devname, devline):
                badlines.append(devline)

    if badlines:
        raise AssertionError(
            "shortid=%s device lines don't contain a "
            "comment with the device name:\n%s" % (osxml.shortid, badlines)
        )


@util.os_parametrize("osxml", filter_devices=True)
def test_devices_duplication(osxml):
    _check_duplicate_devices(osxml)


@util.os_sources_parametrize("osxml", filter_devices=True)
def test_devices_comments(osxml):
    _check_uncommented_devices(osxml)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import locale
from pathlib import Path
import os


def pytest_addoption(parser):
    parser.addoption(
        "--network-tests",
        action="store_true",
        default=False,
        help=(
            "Run osinfo-db network tests. Same tests as triggered "
            "by setting env variable OSINFO_DB_NETWORK_TESTS"
        ),
    )


def pytest_configure(config):
    top = Path(__file__).parent.parent
    key = "INTERNAL_OSINFO_DB_DATA_DIR"
    if key not in os.environ:
        os.environ[key] = str(Path(top, "data").resolve())

    key = "INTERNAL_OSINFO_DB_DATA_SRC_DIR"
    if key not in os.environ:
        os.environ[key] = str(Path(top, "data").resolve())

    # Needed for test reproducibility on any system not using a UTF-8 locale
    locale.setlocale(locale.LC_ALL, "C")
    for loc in ["C.UTF-8", "C.utf8", "UTF-8", "en_US.UTF-8"]:
        try:
            locale.setlocale(locale.LC_CTYPE, loc)
            break
        except locale.Error:
            pass
    else:
        raise locale.Error("No UTF-8 locale found")

    # Default to --log-level=info if not otherwise specified
    if hasattr(config.option, "log_level") and config.option.log_level is None:
        config.option.log_level = "info"

    # This will trigger some DATA_DIR validation
    from . import util

    _ = util


def pytest_ignore_collect(path, config):
    """
    Entirely skip loading test_urls.py if the option wasn't specified
    """
    run_network = bool(
        config.getoption("--network-tests") or os.environ.get("OSINFO_DB_NETWORK_TESTS")
    )
    if Path(path).name == "test_urls.py" and not run_network:
        return True



#!/usr/bin/env python3

# With Nix, run in:
# nix-shell -p "python3.withPackages (p: [ p.lxml p.requests ])" -p cdrkit

"""
This script is supposed to be run when a new stable version of NixOS is released.
It takes the new release number, codename, and release date as inputs and does
the following:

1) Creates new stable osinfo-db entry. This entry is based on the previous entry XML by
   substituting the release numbers and removing unnecessary tags.

2) Updates the unstable entry by bumping the release number to the version that comes
   after the new release, the current in-development version. Also <upgrades> and
   <derives-from> tags are updated to point to the new stable entry.

3) Updates the unknown entry with volume-id regular expression that matches all nixos
   ISOs newer than the unstable release.

4) Downloads ISOs of the new release, runs isoinfo -d on them, and saves the result
   to tests/isodata/nixos.

5) Regenerates the isodata of the unstable release - deletes the old isodata,
   downloads new ISOs, saves the result to tests/isodata/nixos.

Please note the "derives-from" (as well as "upgrades") relation maintains following
chain:

 +-------+      +-------+               +-----------------+      +----------+      +---------+
 | 20.03 | <--- | 20.09 | <--- ... <--- | (latest stable) | <--- | unstable | <--- | unknown |
 +-------+      +-------+               +-----------------+      +----------+      +---------+

When a release is derived from another it inherits its list of supported devices,
firmware types, as well as its logo[1]. The purpose of the unknown entry is to
recognize new NixOS releases on systems with outdated osinfo-db.

It should be safe to run this script multiple times with the same --release argument.
Doctests can be executed with "python -m doctest -v nixos.py". Future archaeologists
might be interested in [2][3]. Good luck!

[1] https://gitlab.gnome.org/GNOME/gnome-boxes/-/blob/master/data/osinfo/nixos-20.03.xml
[2] https://gitlab.com/libosinfo/osinfo-db/-/merge_requests/107
[3] https://github.com/NixOS/nixpkgs/pull/83551

"""

import argparse
import glob
import os.path
import logging
import re
import shutil
import subprocess
import sys

from lxml import etree
import requests

DATA_DIR = os.path.relpath(
    os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "..", "data", "os", "nixos.org")
    )
)

TEST_DIR = os.path.relpath(
    os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "..", "tests", "isodata", "nixos")
    )
)

UNSTABLE = os.path.join(DATA_DIR, "nixos-unstable.xml.in")
UNKNOWN = os.path.join(DATA_DIR, "nixos-unknown.xml.in")
ISO_BYTES = 2 * 1024 * 1024


def fatal(*args):
    logging.critical(*args)
    sys.exit(1)


def guess_next_release(release):
    """
    Guess next release number by adding 6 months.
    >>> guess_next_release("21.09")
    '22.03'
    """
    (y, m) = release.split(".")
    m = int(m) + 6
    y = int(y) + (m / 12)
    m = m % 12
    return "%02d.%02d" % (y, m)


def str_regex(pattern, suggested_format):
    def validate(arg):
        if not re.match(pattern, arg):
            raise argparse.ArgumentTypeError(
                f"invalid format, must be {suggested_format}"
            )
        return arg

    return validate


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--release",
        type=str_regex(r"^\d\d[.]\d\d$", "YY.MM"),
        help="release series number (e.g. 20.03)",
        required=True,
    )
    parser.add_argument(
        "--codename",
        type=str,
        help="release code name (e.g. Markhor)",
        required=True,
    )
    parser.add_argument(
        "--release-date",
        type=str_regex(r"^20\d\d-[01]\d-[0-3]\d$", "YYYY-MM-DD"),
        help="release date (e.g. 2020-04-20)",
        required=True,
    )
    parser.add_argument(
        "--next-release",
        type=str_regex(r"^\d\d[.]\d\d$", "YY.MM"),
        help="next release number after --release (plus 6 months by default)",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="print debugging output",
    )
    args = parser.parse_args()

    if args.next_release is None:
        args.next_release = guess_next_release(args.release)

    return args


def check_isoinfo():
    exe = shutil.which("isoinfo")
    if exe is None:
        fatal("Command 'isoinfo' not in $PATH, please install cdrkit or cdrtools")
    logging.debug("Found isoinfo in %s", exe)


def find_prev_stable(new_release):
    """
    Finds previous stable entry to use as a template for the new release.
    """
    files = glob.glob(DATA_DIR + "/nixos-[0-9][0-9].[0-9][0-9].xml.in")

    # discard the new release from list if it's there to make it
    # possible to run this script multiple times
    files = sorted(filter(lambda f: new_release not in f, files))
    if len(files) < 1:
        fatal("Can't find previous stable release file in %s", DATA_DIR)
    prev_file = files[-1]

    m = re.search(r"nixos-([0-9][0-9][.][0-9][0-9]).xml.in$", prev_file)
    if m is None:
        fatal("Can't parse previous release number: %s", prev_file)

    logging.info(
        "Creating new release %s with %s as a template", new_release, prev_file
    )
    return (m[1], etree.parse(prev_file))


def tag_replace_text(xml, path, text, replacement):
    """
    Performs search and replace on all elements matching given XPath.
    """
    for e in xml.findall(path):
        replaced = str.replace(e.text, text, replacement)
        logging.debug("%s: %s -> %s", path, e.text, replaced)
        e.text = str.replace(e.text, text, replacement)


def set_derives_from(xml_os, release):
    """
    Sets the derives-from and upgrades tags to the given release.
    """
    logging.debug("Setting upgrades, derives-from to %s", release)
    release_id = f"http://nixos.org/nixos/{release}"
    if xml_os.find("upgrades") is not None and xml_os.find("derives-from") is not None:
        xml_os.find("upgrades").set("id", release_id)
        xml_os.find("derives-from").set("id", release_id)
    else:
        # This branch is supposed to be used by 20.09 only and can be removed
        # afterwards.
        sib = xml_os.find("codename")
        e = etree.Element("derives-from", id=release_id)
        sib.addnext(e)
        e = etree.Element("upgrades", id=release_id)
        e.tail = "\n    "
        sib.addnext(e)
        sib.tail = "\n\n    "


def digit_gt(s):
    """
    Helper function that returns regular expression matching all digits greater than s.

    >>> digit_gt("0")
    '[1-9]'
    >>> digit_gt("1")
    '[2-9]'
    >>> digit_gt("2")
    '[3-9]'
    >>> digit_gt("7")
    '[8-9]'
    >>> digit_gt("8")
    '9'
    """
    assert len(s) == 1 and s != "9"
    n = int(s) + 1
    return "9" if n == 9 else "[%s-9]" % n


def regex_higher_than(release):
    """
    Returns regular expression matching all release numbers newer than the function
    argument. Please note that the returned expression also matches release numbers
    that do not have valid month number on the left side of the dot (e.g. 23.66).
    Should not matter, it is only important to not match any release number that are
    older than the argument.

    >>> regex_higher_than("20.03")
    '20.0[4-9]|20.[1-9]\\\\d|2[1-9].\\\\d\\\\d|[3-9]\\\\d.\\\\d\\\\d'
    >>> regex_higher_than("20.09")
    '20.[1-9]\\\\d|2[1-9].\\\\d\\\\d|[3-9]\\\\d.\\\\d\\\\d'
    >>> regex_higher_than("23.03")
    '23.0[4-9]|23.[1-9]\\\\d|2[4-9].\\\\d\\\\d|[3-9]\\\\d.\\\\d\\\\d'
    >>> regex_higher_than("23.09")
    '23.[1-9]\\\\d|2[4-9].\\\\d\\\\d|[3-9]\\\\d.\\\\d\\\\d'
    >>> regex_higher_than("23.11")
    '23.1[2-9]|23.[2-9]\\\\d|2[4-9].\\\\d\\\\d|[3-9]\\\\d.\\\\d\\\\d'
    >>> regex_higher_than("29.09")
    '29.[1-9]\\\\d|[3-9]\\\\d.\\\\d\\\\d'
    >>> regex_higher_than("29.10")
    '29.1[1-9]|29.[2-9]\\\\d|[3-9]\\\\d.\\\\d\\\\d'
    >>> regex_higher_than("29.99")
    '[3-9]\\\\d.\\\\d\\\\d'
    """
    variants = []
    (y, m) = release.split(".")

    # same year
    if m[1] != "9":
        variants.append(y + "." + m[0] + digit_gt(m[1]))
    if m[0] != "9":
        variants.append(y + "." + digit_gt(m[0]) + r"\d")
    # same decade
    if y[1] != "9":
        variants.append(y[0] + digit_gt(y[1]) + r".\d\d")
    # further
    if y[0] != "9":
        variants.append(digit_gt(y[0]) + r"\d.\d\d")

    return "|".join(variants)


def create_new_stable(release, codename, release_date):
    """
    Creates new stable entry based on the previous stable entry. Replaces release
    numbers, sets codename, release date, points derives-from to the previous entry,
    removes redundant tags that are inherited by the derives-from pointer.
    """
    prev_release, xml = find_prev_stable(release)
    xml_os = xml.find("os")
    xml_os.set("id", f"http://nixos.org/nixos/{release}")
    tag_replace_text(xml_os, "short-id", prev_release, release)
    tag_replace_text(xml_os, "name", prev_release, release)
    tag_replace_text(xml_os, "version", prev_release, release)
    tag_replace_text(xml_os, "variant/name", prev_release, release)
    tag_replace_text(xml_os, "media/url", prev_release, release)
    tag_replace_text(xml_os, "media/iso/volume-id", prev_release, release)
    xml_os.find("codename").text = codename
    xml_os.find("release-date").text = release_date

    for tag_to_remove in ("devices", "firmware"):
        for e in xml_os.findall(tag_to_remove):
            logging.debug(
                "Removing <%s> because it is inherited via derives-from", tag_to_remove
            )
            xml_os.remove(e)

    set_derives_from(xml_os, prev_release)
    filename = os.path.join(DATA_DIR, f"nixos-{release}.xml.in")
    xml.write(filename)
    logging.info("Created new stable entry %s", filename)
    return xml


def update_unstable(new_release, next_release):
    """
    Updates the unstable entry to have volume-id matching the next, yet-unreleased,
    version number. Changes derives-from to point to the newly added stable release.
    """
    logging.info("Updating %s to %s", UNSTABLE, next_release)
    xml = etree.parse(UNSTABLE)
    xml_os = xml.find("os")
    if new_release not in xml_os.find("media/iso/volume-id").text:
        logging.warning(
            "Current unstable volume-id does not contain %s - make sure the new value "
            "is correct",
            new_release,
        )
    tag_replace_text(xml_os, "media/iso/volume-id", new_release, next_release)
    set_derives_from(xml_os, new_release)
    xml.write(UNSTABLE)
    logging.info("Updated unstable entry %s", UNSTABLE)
    return xml


def update_unknown(next_release):
    """
    Updates the regular expression of the Unknown NixOS osinfo-db entry. The expression
    should match any version that is higher than what Unstable matches. The purpose
    of the Unknown entry is to make it possible to match NixOS ISOs even on system with
    outdated osinfo-db.
    """
    xml = etree.parse(UNKNOWN)
    regex = regex_higher_than(next_release)
    xml.find("os/media/iso/volume-id").text = "nixos-.*-(" + regex + ")-.*"
    xml.write(UNKNOWN)
    logging.info("Updated unknown entry %s", UNKNOWN)


def fetch_iso(url, directory):
    """
    Downloads first 2 MiB of ISO image to the given directory. As of spring 2020
    the URLs in the entries are redirects pointing to URLs with the ISO filename
    that includes build number and git commit, we make sure to save the ISOs under
    this more descriptive name.
    """
    r = requests.get(url, stream=True)
    r.raise_for_status()
    filename = r.request.url.split("/")[-1]
    logging.debug("%s filename: %s", url, filename)
    filename = os.path.join(directory, filename)
    logging.info("Downloading %s", url)
    with open(filename, "wb") as fh:
        fh.write(next(r.iter_content(chunk_size=ISO_BYTES)))
    logging.debug("Downloaded %s", filename)
    return filename


def run_isoinfo(isofile, outfile):
    """
    Runs isoinfo -d -i isofile and saves it to outfile.
    """
    res = subprocess.run(["isoinfo", "-d", "-i", isofile], capture_output=True)
    if res.returncode != 0:
        logging.error("isoinfo failed on %s:\n%s", isofile, res.stderr)
        return

    with open(outfile, "wb") as fh:
        fh.write(res.stdout)
    logging.info("Saved isoinfo to %s", outfile)


def process_iso(url, directory):
    """
    Download the beginning of ISO image and run isoinfo on it. Errors are not fatal
    to accomodate for scenario where variants are removed.
    """
    try:
        isofile = fetch_iso(url, directory)
    except requests.RequestException as e:
        logging.error("Failed to fetch %s: %s", url, str(e))
        return []
    try:
        testdata_file = isofile + ".txt"
        run_isoinfo(isofile, testdata_file)
    finally:
        os.remove(isofile)
    return [testdata_file]


def update_testdata(new_release, stable_xml, unstable_xml):
    """
    Update testdata used by "make check". Creates isoinfo for the new release.
    Deletes it for the previous unstable release and regenerates it from current
    unstable isos. "Unknown" entry has no testdata.
    """
    add = []
    remove = []

    logging.info("Creating ISO test data for %s", new_release)
    new_release_dir = os.path.join(TEST_DIR, f"nixos-{new_release}")
    os.makedirs(new_release_dir, exist_ok=True)
    for e in stable_xml.findall("os/media/url"):
        add += process_iso(e.text, new_release_dir)

    logging.info("Updating ISO test data for Unstable")
    unstable_dir = os.path.join(TEST_DIR, "nixos-unstable")
    for f in os.listdir(unstable_dir):
        if f.endswith(".iso.txt"):
            logging.warning("Deleting unstable testdata %s", f)
            os.remove(os.path.join(unstable_dir, f))
            remove.append(os.path.join(unstable_dir, f))
    for e in unstable_xml.findall("os/media/url"):
        add += process_iso(e.text, unstable_dir)

    return (add, remove)


def print_instructions(release, git_add, git_rm):
    """
    Remind the user of manual steps that need to be done before submitting merge
    request with the new release.
    """
    new_release_file = os.path.join(DATA_DIR, f"nixos-{release}.xml.in")
    logging.warning(
        "Make sure <variant>s are up to date in %s and %s", new_release_file, UNSTABLE
    )
    logging.warning(
        "Make sure system <resources> are appropriate in %s and %s and %s",
        new_release_file,
        UNSTABLE,
        UNKNOWN,
    )
    logging.warning(
        "If new virtual hardware is supported, add new <devices> to %s",
        new_release_file,
    )
    logging.warning(
        "Run git add:\n%s",
        "\n".join(map(lambda f: f"\tgit add {f}", [new_release_file] + git_add)),
    )
    logging.warning(
        "Run git rm:\n%s", "\n".join(map(lambda f: f"\tgit rm {f}", git_rm))
    )
    logging.warning("Run git commit -a, make check, and go create a merge request!")


def main():
    args = parse_args()
    logging.basicConfig(
        format="%(levelname)s\t%(message)s",
        level=(logging.DEBUG if args.verbose else logging.INFO),
    )
    check_isoinfo()

    stable_xml = create_new_stable(args.release, args.codename, args.release_date)
    unstable_xml = update_unstable(args.release, args.next_release)
    update_unknown(args.next_release)
    add, remove = update_testdata(args.release, stable_xml, unstable_xml)
    print_instructions(args.release, add, remove)


if __name__ == "__main__":
    main()



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from pathlib import Path

from . import util


def _test_validate_ids(xml, entity_type):
    """
    Ensure the ids are the ones supported by OsinfoLoader.

    This check tries to mimic, in a pythonic way, the very same
    check done by OsinfoLoader::osinfo_loader_check_id()
    """
    base = Path("data", entity_type)
    relpath = xml.path.relative_to(base)

    suffix = xml.internal_id[len("http://") :]
    vendor = suffix.split("/", 1)[0]
    entity_name = suffix.split("/", 1)[1].replace("/", "-")

    if relpath.parent.suffix == ".d":
        filename = vendor + "/" + entity_name + ".d"
        assert filename == str(relpath.parent)
    else:
        filename = vendor + "/" + entity_name + ".xml"
        assert filename == str(relpath)


@util.os_parametrize("osxml")
def test_validate_os_ids(osxml):
    """
    Ensure the OS ids are the ones supported by OsinfoLoader.
    """
    return _test_validate_ids(osxml, "os")


@util.device_parametrize("devicexml")
def test_validate_device_ids(devicexml):
    """
    Ensure the Device ids are the ones supported by OsinfoLoader.
    """
    return _test_validate_ids(devicexml, "device")


@util.datamap_parametrize("datamapxml")
def test_validate_datamap_ids(datamapxml):
    """
    Ensure the Datamap ids are the ones supported by OsinfoLoader.
    """
    return _test_validate_ids(datamapxml, "datamap")


@util.installscript_parametrize("installscriptxml")
def test_validate_installscript_ids(installscriptxml):
    """
    Ensure the InstallScript ids are the ones supported by OsinfoLoader.
    """
    return _test_validate_ids(installscriptxml, "install-script")


@util.platform_parametrize("platformxml")
def test_validate_platform_ids(platformxml):
    """
    Ensure the Platform ids are the ones supported by OsinfoLoader.
    """
    return _test_validate_ids(platformxml, "platform")



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import configparser


class _TreeinfoData:
    def __init__(self, path, arch, family, variant, version):

        self.path = path
        self.arch = arch or ""
        self.family = family or ""
        self.variant = variant or ""
        self.version = version or ""

    def match(self, treeinfo):
        if (
            bool(treeinfo.arch.match(self.arch))
            and bool(treeinfo.family.match(self.family))
            and bool(treeinfo.variant.match(self.variant))
            and bool(treeinfo.version.match(self.version))
        ):
            return True

        return False


def get_treeinfodata(filepath):
    arch = None
    family = None
    variant = None
    version = None

    config = configparser.ConfigParser()
    with filepath.open("r") as out:
        config.read_file(out)
        arch = ""
        family = ""
        variant = ""
        version = ""

        if "arch" in config["general"]:
            arch = config["general"]["arch"]

        if "family" in config["general"]:
            family = config["general"]["family"]

        if "variant" in config["general"]:
            variant = config["general"]["variant"]

        if "version" in config["general"]:
            version = config["general"]["version"]

    return _TreeinfoData(filepath, arch, family, variant, version)



#!/usr/bin/env python3

import argparse
import distutils.spawn
import os
import sys
import tempfile
import time


topdir = os.path.realpath(os.path.join(os.path.dirname(__file__), ".."))
datadir = os.path.join(topdir, "data")
sys.path.insert(0, topdir)
os.environ["INTERNAL_OSINFO_DB_DATA_DIR"] = datadir
os.environ["INTERNAL_OSINFO_DB_DATA_SRC_DIR"] = datadir

import tests.isodata
import tests.util


def fail(msg):
    print(msg)
    sys.exit(1)


##############################
# main() and option handling #
##############################


def _parse_args():
    desc = (
        "Helper script for adding iso test data to the test "
        "suite, and matching <os> <media> data to the DB"
    )
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument("shortid", help="Which <os> short-id the ISO is media for")
    parser.add_argument("iso", help="The path to the ISO media")
    parser.add_argument(
        "--arch",
        default="x86_64",
        help="The OS architecture the media is for. default=x86_64",
    )

    options = parser.parse_args()
    return options


def _main():
    """
    This is a template for new command line programs. Copy and edit it!
    """
    options = _parse_args()

    iso = os.path.realpath(os.path.abspath(options.iso))
    isoinfobin = distutils.spawn.find_executable("isoinfo")
    if not os.path.exists(iso):
        fail("iso does not exist: %s" % iso)
    if not isoinfobin:
        fail("isoinfo is not installed")

    osxml = None
    for o in tests.util.DataFiles.oses():
        if o.shortid == options.shortid:
            osxml = o
            break
    if not osxml:
        fail("Did not find any os shortid=%s" % options.shortid)
        return

    destdir = os.path.join(topdir, "tests", "isodata", osxml.distro, options.shortid)
    destpath = os.path.join(destdir, os.path.basename(iso) + ".txt")

    tmp = tempfile.NamedTemporaryFile()
    ret = os.system("isoinfo -d -i %s > %s" % (iso, tmp.name))
    if ret != 0:
        fail("Command failed, returncode=%s" % ret)

    # parse isoinfo
    # output an example media block
    isodata = tests.isodata.get_isodatamedia(tmp.name)
    print("XML to add to %s :" % osxml.filename[len(topdir) + 1 :] + ".in")
    print()
    print('    <media arch="%s">' % options.arch)
    print("      <url>XXX</url>")
    print("      <iso>")

    if isodata.volumeid:
        print("        <volume-id>%s</volume-id>" % isodata.volumeid)
    if isodata.systemid:
        print("        <system-id>%s</system-id>" % isodata.systemid)
    if isodata.publisherid:
        print("        <publisher-id>%s</publisher-id>" % isodata.publisherid)
    if isodata.applicationid:
        print("        <application-id>%s</application-id>" % isodata.applicationid)
    if isodata.volumesize:
        print("        <volume-size>%s</volume-size>" % isodata.volumesize)

    print("      </iso>")
    print("      <kernel>XXX</kernel>")
    print("      <initrd>XXX</initrd>")
    print("    </media>")
    print()

    print("\n\nSleeping 5 seconds before writing test data to:\n%s" % destpath)
    time.sleep(5)
    if not os.path.exists(destdir):
        os.system("mkdir -p %s" % destdir)
    os.system("cp %s %s" % (tmp.name, destpath))
    print("Done.")

    return 0


if __name__ == "__main__":
    sys.exit(_main())



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import util


def _check_duplicate_firmwares(osxml):
    """
    Ensure an OS doesn't list a firmware that's defined in the parent
    """
    broken = []
    related = util.DataFiles.getosxml_related(osxml)
    for osxml2 in related:
        if osxml2.firmwares is not None:
            for firmware2 in osxml2.firmwares:
                for firmware in osxml.firmwares:
                    if (
                        firmware.arch == firmware2.arch
                        and firmware.firmware_type == firmware2.firmware_type
                    ):
                        broken.append([firmware.firmware_type, firmware.arch])
    assert broken == []


@util.os_parametrize("osxml", filter_firmwares=True)
def test_firmwares_duplication(osxml):
    _check_duplicate_firmwares(osxml)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from collections import defaultdict

import logging

from . import util


def _test_resources_uniqueness_by_arch(osxml):
    """Ensure there's no more than one resource element per architecture"""
    result = defaultdict(list)
    for resources in osxml.resources_list:
        result[resources.get("arch")].append(resources)

    for value in result.values():
        assert len(value) == 1


@util.os_parametrize("osxml", filter_resources=True)
def test_resources(osxml):
    _test_resources_uniqueness_by_arch(osxml)

    # Ensure minimum resources are <= recommended resources
    _resources_helper(
        osxml,
        osxml.get_minimum_resources,
        "minimum",
        osxml.get_recommended_resources,
        "recommended",
    )

    # Ensure minimum resources are <= maximum resources
    _resources_helper(
        osxml,
        osxml.get_minimum_resources,
        "minimum",
        osxml.get_recommended_resources,
        "maximum",
    )

    # Ensure recommended resources are <= maximum resources
    _resources_helper(
        osxml,
        osxml.get_recommended_resources,
        "recommended",
        osxml.get_maximum_resources,
        "maximum",
    )

    # Ensure minimum resources <= network resources
    _resources_helper(
        osxml,
        osxml.get_recommended_resources,
        "minimum",
        osxml.get_network_install_resources,
        "network-install",
    )

    # Ensure recommended resources <= network resources
    _resources_helper(
        osxml,
        osxml.get_recommended_resources,
        "recommended",
        osxml.get_network_install_resources,
        "network-install",
    )

    # Ensure network resources <= maximum resources
    _resources_helper(
        osxml,
        osxml.get_network_install_resources,
        "network-install",
        osxml.get_maximum_resources,
        "maximum",
    )


def _resources_helper(osxml, smaller_func, smaller_str, bigger_func, bigger_str):
    broken = []
    for resource in osxml.resources_list:
        logging.info("resources | arch: %s", resource.get("arch"))
        smaller = smaller_func(resource)
        bigger = bigger_func(resource)

        if smaller is None or bigger is None:
            continue

        if not _resources_check(smaller, smaller_str, bigger, bigger_str):
            broken.append([smaller, bigger])
    assert broken == []


def _resources_check(smaller, smaller_str, bigger, bigger_str):
    ret = True
    if smaller.cpu is not None and bigger.cpu is not None:
        if smaller.cpu > bigger.cpu:
            logging.warning(
                "cpu value of %s should not be bigger than %s " "('%d' > '%d')",
                smaller_str,
                bigger_str,
                smaller.cpu,
                bigger.cpu,
            )
            ret = False
    if smaller.n_cpus is not None and bigger.n_cpus is not None:
        if smaller.n_cpus > bigger.n_cpus:
            logging.warning(
                "n-cpus value of %s should not be bigger than %s " "('%d' > '%d')",
                smaller_str,
                bigger_str,
                smaller.n_cpus,
                bigger.n_cpus,
            )
            ret = False
    if smaller.ram is not None and bigger.ram is not None:
        if smaller.ram > bigger.ram:
            logging.warning(
                "ram value of %s should not be bigger than %s " "('%d' > '%d')",
                smaller_str,
                bigger_str,
                smaller.ram,
                bigger.ram,
            )
            ret = False
    if smaller.storage is not None and bigger.storage is not None:
        if smaller.storage > bigger.storage:
            logging.warning(
                "storage value of %s should not be bigger than %s " "('%d' > '%d')",
                smaller_str,
                bigger_str,
                smaller.storage,
                bigger.storage,
            )
            ret = False
    return ret



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import logging
from pathlib import Path
import pytest

from . import util
from . import treeinfodata


def _get_treeinfodatapaths():
    """
    Collect treeinfo data and return a list of tuples:
        (osname, treeinfodatapaths)
    """
    this = Path(__file__).resolve()
    treeinfodata_path = Path(this.parent, "treeinfodata")

    ret = []
    allpaths = treeinfodata_path.glob(str(Path("*", "*")))
    for osdir in sorted(allpaths, key=util.path_sort):
        # The .treeinfo may be under the following directories schema:
        # - distro/shortid/arch
        # - distro/shortid/variant/arch/
        # That's the reason we have to take check for the .treeinfo file recursively.
        treeinfodatapaths = list(osdir.rglob("*.treeinfo"))
        if len(treeinfodatapaths):
            ret.append((osdir.name, treeinfodatapaths))
    return ret


@pytest.mark.parametrize("testdata", _get_treeinfodatapaths(), ids=lambda d: d[0])
def test_treeinfo_detection(testdata):
    osname, treeinfodatapaths = testdata
    for treeinfodatapath in treeinfodatapaths:
        detected = []
        data = treeinfodata.get_treeinfodata(treeinfodatapath)

        for osxml2 in util.DataFiles.oses():
            for tree in osxml2.trees:
                if tree.treeinfo is None:
                    continue

                if data.match(tree.treeinfo):
                    if osname != osxml2.shortid:
                        logging.warning(
                            "treeinfo '%s' was matched by OS '%s' while "
                            "it should only be matched by OS '%s'",
                            data.path,
                            osxml2.shortid,
                            osname,
                        )
                    else:
                        logging.info(
                            "treeinfo '%s' matched by OS '%s'",
                            data.path,
                            osxml2.shortid,
                        )

                    if osxml2.shortid not in detected:
                        detected.append(osxml2.shortid)

        if detected == [osname]:
            continue

        raise AssertionError(
            "data: %s\nMatched=%s but expected=%s"
            % (treeinfodatapath, detected, [osname])
        )



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import re

import lxml.etree as ET


def _cache_property(fn):
    """
    Decorator to use self._cache to cache property lookup results
    """

    def _wrapper(*args):
        self = args[0]
        key = str(fn)
        cache = self._cache  # pylint: disable=protected-access
        if key not in cache:
            cache[key] = fn(*args)
        return cache[key]

    return property(_wrapper)


class _XMLBase:
    """
    Simple base class for our XML parsers
    """

    def __init__(self, root):
        self._root = root
        self._cache = {}

    def _get_text(self, element_name, default=None):
        node = self._root.find(element_name)
        if node is not None:
            return node.text
        return default

    def _get_int(self, element_name, default=None):
        text = self._get_text(element_name)
        if text is not None:
            return int(text)
        return default


class Os(_XMLBase):
    def __init__(self, path):
        self.path = path
        root = ET.parse(self.path.open("r")).getroot().find("os")
        super().__init__(root)

    def __repr__(self):
        return "<%s shortid=%s>" % (self.__class__.__name__, self.shortid)

    @_cache_property
    def internal_id(self):
        return self._root.get("id")

    @_cache_property
    def derives_from(self):
        derives_from = self._root.find("derives-from")
        if derives_from is not None:
            return derives_from.get("id")
        return None

    @_cache_property
    def clones(self):
        clones = self._root.find("clones")
        if clones is not None:
            return clones.get("id")
        return None

    @_cache_property
    def upgrades(self):
        upgrades = self._root.find("upgrades")
        if upgrades is not None:
            return upgrades.get("id")
        return None

    @_cache_property
    def devices(self):
        devices = []
        devicelist = self._root.find("devices")
        if devicelist is not None:
            for device in devicelist.findall("device"):
                if device.get("supported") != "false":
                    devices.append(device.get("id"))
        return devices

    @_cache_property
    def devices_unsupported(self):
        devices = []
        devicelist = self._root.find("devices")
        if devicelist is not None:
            for device in devicelist.findall("device"):
                if device.get("supported") == "false":
                    devices.append(device.get("id"))
        return devices

    @_cache_property
    def images(self):
        images = []
        for image in self._root.findall("image"):
            images.append(Image(image))
        return images

    @_cache_property
    def medias(self):
        medias = []
        for media in self._root.findall("media"):
            medias.append(Media(media))
        return medias

    @_cache_property
    def trees(self):
        trees = []
        for tree in self._root.findall("tree"):
            trees.append(Tree(tree))
        return trees

    @_cache_property
    def shortid(self):
        return self._get_text("short-id")

    @_cache_property
    def shortids(self):
        shortids = []
        for shortid in self._root.findall("short-id"):
            shortids.append(shortid.text)
        return shortids

    @_cache_property
    def distro(self):
        return self._get_text("distro")

    @_cache_property
    def resources_list(self):
        return self._root.findall("resources")

    def _get_resources(self, node, resource_type):
        valid_resources = ["minimum", "recommended", "maximum", "network-install"]
        if resource_type not in valid_resources:
            return None
        # pylint: disable=unsupported-membership-test
        if node not in self.resources_list:
            return None
        resource = node.find(resource_type)
        if resource is not None:
            return Resources(resource)
        return None

    def get_minimum_resources(self, node):
        return self._get_resources(node, "minimum")

    def get_recommended_resources(self, node):
        return self._get_resources(node, "recommended")

    def get_maximum_resources(self, node):
        return self._get_resources(node, "maximum")

    def get_network_install_resources(self, node):
        return self._get_resources(node, "network-install")

    @_cache_property
    def release_date(self):
        return self._get_text("release-date")

    @_cache_property
    def eol_date(self):
        return self._get_text("eol-date")

    @_cache_property
    def variants(self):
        return [v.attrib["id"] for v in self._root.findall("variant")]

    @_cache_property
    def firmwares(self):
        firmwares = []
        firmwarelist = self._root.findall("firmware")
        for firmware in firmwarelist:
            firmwares.append(Firmware(firmware))
        return firmwares

    @_cache_property
    def installscripts(self):
        return [
            s.attrib["id"]
            for i in self._root.findall("installer")
            for s in i.findall("script")
        ]


class Resources(_XMLBase):
    @_cache_property
    def cpu(self):
        return self._get_int("cpu")

    @_cache_property
    def n_cpus(self):
        return self._get_int("n-cpus")

    @_cache_property
    def ram(self):
        return self._get_int("ram")

    @_cache_property
    def storage(self):
        return self._get_int("storage")


class Firmware(_XMLBase):
    @_cache_property
    def arch(self):
        return self._root.attrib["arch"]

    @_cache_property
    def firmware_type(self):
        return self._root.attrib["type"]


class Image(_XMLBase):
    @_cache_property
    def url(self):
        return self._get_text("url")

    @_cache_property
    def variant(self):
        variant = self._root.find("variant")
        if variant is not None:
            return variant.attrib["id"]
        return None

    @_cache_property
    def format(self):
        return self._root.attrib.get("format", None)


class Media(_XMLBase):
    @_cache_property
    def url(self):
        return self._get_text("url")

    @_cache_property
    def iso(self):
        iso = self._root.find("iso")
        if iso is not None:
            return ISO(iso)
        return None

    @_cache_property
    def variant(self):
        variant = self._root.find("variant")
        if variant is not None:
            return variant.attrib["id"]
        return None

    @_cache_property
    def installscripts(self):
        return [
            s.attrib["id"]
            for i in self._root.findall("installer")
            for s in i.findall("script")
        ]


class Tree(_XMLBase):
    @_cache_property
    def url(self):
        return self._get_text("url")

    @_cache_property
    def variant(self):
        variant = self._root.find("variant")
        if variant is not None:
            return variant.attrib["id"]
        return None

    @_cache_property
    def treeinfo(self):
        treeinfo = self._root.find("treeinfo")
        if treeinfo is not None:
            return Treeinfo(treeinfo)
        return None

    @_cache_property
    def kernel(self):
        return self._get_text("kernel")

    @_cache_property
    def initrd(self):
        return self._get_text("initrd")


class Treeinfo(_XMLBase):
    @_cache_property
    def arch(self):
        return re.compile(self._get_text("arch", default=""))

    @_cache_property
    def family(self):
        return re.compile(self._get_text("family", default=""))

    @_cache_property
    def variant(self):
        return re.compile(self._get_text("variant", default=""))

    @_cache_property
    def version(self):
        return re.compile(self._get_text("version", default=""))


class ISO(_XMLBase):
    @_cache_property
    def volumeid(self):
        return re.compile(self._get_text("volume-id", default=""))

    @_cache_property
    def publisherid(self):
        return re.compile(self._get_text("publisher-id", default=""))

    @_cache_property
    def applicationid(self):
        return re.compile(self._get_text("application-id", default=""))

    @_cache_property
    def systemid(self):
        return re.compile(self._get_text("system-id", default=""))

    @_cache_property
    def volumesize(self):
        return self._get_int("volume-size", default=0)


class Device(_XMLBase):
    def __init__(self, path):
        self.path = path
        root = ET.parse(self.path.open("r")).getroot().find("device")
        super().__init__(root)

    @_cache_property
    def internal_id(self):
        return self._root.attrib["id"]

    @_cache_property
    def name(self):
        return self._get_text("name")


class Datamap(_XMLBase):
    def __init__(self, path):
        self.path = path
        root = ET.parse(self.path.open("r")).getroot().find("datamap")
        super().__init__(root)

    def __repr__(self):
        return "<%s path=%s>" % (self.__class__.__name__, self.path)

    @_cache_property
    def internal_id(self):
        return self._root.get("id")


class InstallScript(_XMLBase):
    def __init__(self, path):
        self.path = path
        root = ET.parse(self.path.open("r")).getroot().find("install-script")
        super().__init__(root)

    def __repr__(self):
        return "<%s path=%s>" % (self.__class__.__name__, self.path)

    @_cache_property
    def internal_id(self):
        return self._root.get("id")


class Platform(_XMLBase):
    def __init__(self, path):
        self.path = path
        root = ET.parse(self.path.open("r")).getroot().find("platform")
        super().__init__(root)

    def __repr__(self):
        return "<%s path=%s>" % (self.__class__.__name__, self.path)

    @_cache_property
    def internal_id(self):
        return self._root.get("id")



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from collections import defaultdict

import os
from pathlib import Path
import re

import pytest

from . import osinfo


def human_sort(text):
    # natural/human sorting
    # https://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside
    def atof(t):
        try:
            retval = float(t)
        except ValueError:
            retval = t
        return retval

    return [atof(c) for c in re.split(r"[+-]?([0-9]+(?:[.][0-9]*)?|[.][0-9]+)", text)]


def path_sort(path):
    return human_sort(str(path))


class _Files:
    """
    Track a list of DATA_DIR files and provide APIs for querying them.
    """

    def __init__(self, dir_env, files_format):
        self.datadir = Path(os.environ[dir_env])
        self.schema = Path(self.datadir, "schema", "osinfo.rng")
        self._all_xml_cache = []
        self._oses_cache = []
        self._devices_cache = []
        self._datamaps_cache = []
        self._installscripts_cache = {}
        self._platforms_cache = []
        self._firmwares_cache = []
        self._os_related_cache = defaultdict(list)
        self._files_format = files_format

        if not self.datadir.exists():
            raise RuntimeError("%s=%s doesn't exist" % (dir_env, self.datadir))

    def _get_all_xml(self):
        """
        Get and cache the full list of all DATA_DIR .xml paths
        """
        if not self._all_xml_cache:
            for path in sorted(
                self.datadir.rglob("*" + self._files_format), key=path_sort
            ):
                self._all_xml_cache.append(path)
        return self._all_xml_cache

    def _filter_xml(self, dirname):
        """
        Filter XML paths by those in $DATA_DIR/$dirname
        """

        # Copy of pathlib.is_relative_to, new in 3.9
        def is_relative_to(this, that):
            try:
                this.relative_to(that)
                return True
            except ValueError:
                return False

        return [
            p
            for p in self._get_all_xml()
            if is_relative_to(p, Path(self.datadir, dirname))
        ]

    def oses(
        self,
        filter_media=False,
        filter_trees=False,
        filter_images=False,
        filter_devices=False,
        filter_resources=False,
        filter_dates=False,
        filter_related=False,
        filter_firmwares=False,
        filter_installscripts=False,
    ):
        """
        Return a list of osinfo.Os objects

        :param filter_FOO: Only return objects that have at least one
            instance of a FOO object
        """
        if not self._oses_cache:
            for path in self._filter_xml("os"):
                self._oses_cache.append(osinfo.Os(Path(path)))

        oses = self._oses_cache[:]
        if filter_media:
            oses = [o for o in oses if o.medias]
        if filter_trees:
            oses = [o for o in oses if o.trees]
        if filter_images:
            oses = [o for o in oses if o.images]
        if filter_devices:
            oses = [o for o in oses if o.devices]
        if filter_resources:
            oses = [o for o in oses if o.resources_list]
        if filter_dates:
            oses = [o for o in oses if o.release_date or o.eol_date]
        if filter_related:
            oses = [o for o in oses if self.getosxml_related(o)]
        if filter_firmwares:
            oses = [o for o in oses if o.firmwares]
        if filter_installscripts:
            oses = [o for o in oses if o.installscripts]
        return oses

    def getosxml_by_id(self, os_id):
        for osxml in self.oses():
            if os_id == osxml.internal_id:
                return osxml
        return None

    def getosxml_related(self, osxml):
        if osxml.internal_id not in self._os_related_cache:
            directly_related = []
            if osxml.derives_from is not None:
                found = self.getosxml_by_id(osxml.derives_from)
                assert found
                directly_related.append(found)

            if osxml.clones is not None:
                found = self.getosxml_by_id(osxml.clones)
                assert found
                directly_related.append(found)

            self._os_related_cache[osxml.internal_id].extend(directly_related)

            related = []
            for osxml2 in directly_related:
                related.extend(self.getosxml_related(osxml2))

            for osxml2 in related:
                if osxml2 not in self._os_related_cache[osxml.internal_id]:
                    self._os_related_cache[osxml.internal_id].append(osxml2)
        return self._os_related_cache[osxml.internal_id]

    def devices(self):
        if not self._devices_cache:
            for path in self._filter_xml("device"):
                self._devices_cache.append(osinfo.Device(Path(path)))
        return self._devices_cache

    def datamaps(self):
        if not self._datamaps_cache:
            for path in self._filter_xml("datamap"):
                self._datamaps_cache.append(osinfo.Datamap(Path(path)))
        return self._datamaps_cache

    def installscripts(self):
        if not self._installscripts_cache:
            for path in self._filter_xml("install-script"):
                script = osinfo.InstallScript(Path(path))
                self._installscripts_cache[script.internal_id] = script
        return self._installscripts_cache

    def platforms(self):
        if not self._platforms_cache:
            for path in self._filter_xml("platform"):
                self._platforms_cache.append(osinfo.Platform(Path(path)))
        return self._platforms_cache

    def firmwares(self):
        if not self._firmwares_cache:
            for path in self._filter_xml("firmware"):
                self._firmwares_cache.append(osinfo.Firmware(Path(path)))
        return self._firmwares_cache

    def xmls(self):
        return [Path(p) for p in self._get_all_xml()]


class _DataFiles(_Files):
    """
    Track a list of DATA_DIR XML files and provide APIs for querying
    them. Meant to be initialized only once
    """

    def __init__(self):
        _Files.__init__(self, "INTERNAL_OSINFO_DB_DATA_DIR", ".xml")


DataFiles = _DataFiles()


def _generic_ids_cb(obj, key):
    # pytest passes us a weird value when the list of entities is empty,
    # which might happen depending on how agressively we filter. So
    # we can't assume we are passed a specific entity instance.
    return getattr(obj, key, str(obj))


def _shortid_ids_cb(xml):
    return _generic_ids_cb(xml, "shortid")


def _filename_ids_cb(xml):
    return _generic_ids_cb(xml, "filename")


def os_parametrize(argname, **kwargs):
    """
    Helper for parametrizing a test with an OS list. Passthrough any
    extra arguments to DataFiles.oses()
    """
    oses = DataFiles.oses(**kwargs)
    return pytest.mark.parametrize(argname, oses, ids=_shortid_ids_cb)


def device_parametrize(argname, **kwargs):
    """
    Helper for parametrizing a test with an OS list. Passthrough any
    extra arguments to DataFiles.oses()
    """
    devices = DataFiles.devices(**kwargs)
    return pytest.mark.parametrize(argname, devices, ids=_filename_ids_cb)


def datamap_parametrize(argname, **kwargs):
    """
    Helper for parametrizing a test with an OS list. Passthrough any
    extra arguments to DataFiles.oses()
    """
    datamaps = DataFiles.datamaps(**kwargs)
    return pytest.mark.parametrize(argname, datamaps, ids=_filename_ids_cb)


def installscript_parametrize(argname, **kwargs):
    """
    Helper for parametrizing a test with an OS list. Passthrough any
    extra arguments to DataFiles.oses()
    """
    installscripts = DataFiles.installscripts(**kwargs).values()
    return pytest.mark.parametrize(argname, installscripts, ids=_filename_ids_cb)


def platform_parametrize(argname, **kwargs):
    """
    Helper for parametrizing a test with an OS list. Passthrough any
    extra arguments to DataFiles.oses()
    """
    platforms = DataFiles.platforms(**kwargs)
    return pytest.mark.parametrize(argname, platforms, ids=_filename_ids_cb)


class _SourceFiles(_Files):
    """
    Track a list of DATA_SRC_DIR XML.IN files and provide APIs for querying
    them. Meant to be initialized only once
    """

    def __init__(self):
        _Files.__init__(self, "INTERNAL_OSINFO_DB_DATA_SRC_DIR", ".xml.in")


SourceFiles = _SourceFiles()


def os_sources_parametrize(argname, **kwargs):
    """
    Helper for parametrizing a test with an OS list. Passthrough any
    extra arguments to DataFiles.oses()
    """

    oses = SourceFiles.oses(**kwargs)
    return pytest.mark.parametrize(argname, oses, ids=_shortid_ids_cb)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import util


@util.os_parametrize("osxml", filter_installscripts=True)
def test_install_scripts(osxml):
    for script in osxml.installscripts:
        assert util.DataFiles.installscripts()[script]


@util.os_parametrize("osxml", filter_media=True)
def test_media_install_scripts(osxml):
    for media in osxml.medias:
        for script in media.installscripts:
            assert util.DataFiles.installscripts()[script]



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import enum
import http
import logging

import pytest
import requests

from . import util


class UrlType(enum.Enum):
    URL_GENERIC = 1
    URL_ISO = 2
    URL_INITRD = 3
    URL_DISK_RAW = 4
    URL_DISK_QCOW2 = 5
    URL_DISK_VMDK = 6
    URL_TREEINFO = 7
    URL_DISK_CONTAINERDISK = 8


iso_content_types = {
    # proper ISO mimetype
    "application/x-cd-image",
    "application/x-iso9660-image",
    # generic data
    "application/octet-stream",
    "binary/octet-stream",
    # ISO files on archive.netbsd.org
    "text/plain",
    # a few openSUSE Live images
    "application/x-up-download",
}


initrd_content_types = {
    # generic data
    "application/octet-stream",
    # gzip-compressed
    "application/x-gzip",
}


raw_content_types = {}


qcow2_content_types = {
    # generic data
    "application/octet-stream",
    # qcow2 files on fedoraproject.org mirrors; similar issue of
    # https://pagure.io/fedora-infrastructure/issue/10766
    "application/x-troff-man",
    # qcow2 files on some opensuse.org mirrors
    "text/plain",
}


vmdk_content_types = {}


treeinfo_content_types = {
    # generic data
    "application/octet-stream",
}


containerdisk_content_types = {
    # image manifest
    "application/vnd.docker.distribution.manifest.v1+json",
}


def _is_content_type_allowed(content_type, url_type):
    if url_type == UrlType.URL_ISO:
        return content_type in iso_content_types
    if url_type == UrlType.URL_INITRD:
        return content_type in initrd_content_types
    if url_type == UrlType.URL_DISK_RAW:
        return content_type in raw_content_types
    if url_type == UrlType.URL_DISK_QCOW2:
        return content_type in qcow2_content_types
    if url_type == UrlType.URL_DISK_VMDK:
        return content_type in vmdk_content_types
    if url_type == UrlType.URL_TREEINFO:
        return content_type in treeinfo_content_types
    if url_type == UrlType.URL_DISK_CONTAINERDISK:
        return content_type in containerdisk_content_types
    return True


def _transform_docker_url(url):
    """
    Transform docker:// url into a docker registry API call
    See: https://docs.docker.com/registry/spec/api/#existing-manifests
    """
    url_parts = url.split("/")
    url = f"http://{url_parts[2]}/v2/"
    for i in range(3, len(url_parts) - 1):
        url += f"{url_parts[i]}/"
    image, tag = url_parts[-1].split(":")
    url += f"{image}/manifests/{tag}"
    return url


def _check_url(url, url_type):
    logging.info("url: %s, type: %s", url, url_type)
    headers = {"user-agent": "Wget/1.0"}
    if url_type == UrlType.URL_DISK_CONTAINERDISK:
        url = _transform_docker_url(url)
    response = requests.head(url, allow_redirects=True, headers=headers, timeout=30)
    content_type = response.headers.get("content-type")
    if content_type:
        try:
            content_type = content_type[0 : content_type.index(";")]
        except ValueError:
            pass
    logging.info(
        "response: %s; code: %d; content-type: %s",
        http.client.responses[response.status_code],
        response.status_code,
        content_type,
    )
    if not response.ok:
        return False
    if content_type and not _is_content_type_allowed(content_type, url_type):
        return False
    return True


def _collect_os_urls():
    """
    Iterate the OS list and return a list of pairs (shortid, [url list)
    """
    ret = []

    for osxml in util.DataFiles.oses():
        urls = []
        for i in osxml.images:
            if not i.url:
                continue
            url_type = UrlType.URL_GENERIC
            if i.format == "raw":
                url_type = UrlType.URL_DISK_RAW
            elif i.format == "qcow2":
                url_type = UrlType.URL_DISK_QCOW2
            elif i.format == "vmdk":
                url_type = UrlType.URL_DISK_VMDK
            elif i.format == "containerdisk":
                url_type = UrlType.URL_DISK_CONTAINERDISK
            urls.append((i.url, url_type))
        urls.extend([(m.url, UrlType.URL_ISO) for m in osxml.medias if m.url])
        for t in osxml.trees:
            if not t.url:
                continue
            urls.append((t.url, UrlType.URL_GENERIC))
            url = t.url
            if not url.endswith("/"):
                url += "/"
            if t.kernel:
                urls.append((url + t.kernel, UrlType.URL_GENERIC))
            if t.initrd:
                urls.append((url + t.initrd, UrlType.URL_INITRD))
            if t.treeinfo:
                urls.append((url + ".treeinfo", UrlType.URL_TREEINFO))
        if urls:
            ret.append((osxml.shortid, urls))

    return ret


@pytest.mark.parametrize(
    "testdata", _collect_os_urls(), ids=lambda testdata: testdata[0]
)
def test_urls(testdata):
    urls = testdata[1]
    broken = []
    for (url, url_type) in urls:
        # As some distro URLs are flaky, let's give it a try 3 times
        # before actually failing.
        for i in range(3):
            ok = _check_url(url, url_type)
            if ok:
                break

        if not ok:
            broken.append(url)
    assert broken == []



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.


class _ISODataMedia:
    def __init__(
        self, path, volumeid, publisherid, systemid, applicationid, volumesize
    ):

        self.path = path
        self.volumeid = volumeid if volumeid is not None else ""
        self.publisherid = publisherid if publisherid is not None else ""
        self.systemid = systemid if systemid is not None else ""
        self.applicationid = applicationid if applicationid is not None else ""
        self.volumesize = volumesize if volumesize is not None else 0

    def match(self, media):
        volumesize = media.volumesize
        if volumesize == 0:
            volumesize = self.volumesize

        if (
            bool(media.volumeid.match(self.volumeid))
            and bool(media.publisherid.match(self.publisherid))
            and bool(media.applicationid.match(self.applicationid))
            and bool(media.systemid.match(self.systemid))
            and volumesize == self.volumesize
        ):
            return True

        return False


def _get_value(string, prefix, return_type=str):
    if string.startswith(prefix):
        return return_type(string[len(prefix) :].strip())
    return None


def _get_volumeid(string):
    return _get_value(string, "Volume id: ")


def _get_publisherid(string):
    return _get_value(string, "Publisher id: ")


def _get_systemid(string):
    return _get_value(string, "System id: ")


def _get_applicationid(string):
    return _get_value(string, "Application id: ")


def _get_logicalblock(string):
    return _get_value(string, "Logical block size is: ", int)


def _get_volumesize(string):
    return _get_value(string, "Volume size is: ", int)


def get_isodatamedia(filepath):
    volumeid = None
    publisherid = None
    systemid = None
    applicationid = None
    logicalblock = None
    volumesize = None

    with filepath.open("r") as out:
        for line in out.readlines():
            if volumeid is None:
                volumeid = _get_volumeid(line)
            if publisherid is None:
                publisherid = _get_publisherid(line)
            if systemid is None:
                systemid = _get_systemid(line)
            if applicationid is None:
                applicationid = _get_applicationid(line)
            if logicalblock is None:
                logicalblock = _get_logicalblock(line)
            if volumesize is None:
                volumesize = _get_volumesize(line)

    if logicalblock is not None and volumesize is not None:
        volumesize *= logicalblock
    else:
        volumesize = None

    return _ISODataMedia(
        filepath, volumeid, publisherid, systemid, applicationid, volumesize
    )



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import lxml.etree as ET
import pytest

from . import util


SCHEMA = util.DataFiles.schema
RELAXNG = ET.RelaxNG(ET.parse(SCHEMA.open("r")))


@pytest.mark.parametrize("path", util.DataFiles.xmls())
def test_schema(path):
    if not RELAXNG.validate(ET.parse(path.open("r"))):
        # pylint: disable=no-member
        raise AssertionError(str(RELAXNG.error_log.last_error))



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import logging
from pathlib import Path
import pytest

from . import util
from . import isodata


def _get_isodatapaths():
    """
    Collect iso media data and return a list of tuples:
        (osname, isodatapaths)
    """
    this = Path(__file__).resolve()
    isodata_path = Path(this.parent, "isodata")

    ret = []
    osdirs = [d for d in isodata_path.glob(str(Path("*", "*"))) if d.is_dir()]

    for osdir in sorted(osdirs, key=util.path_sort):
        isodatapaths = list(osdir.glob("*.txt"))
        if len(isodatapaths):
            ret.append((osdir.name, isodatapaths))

    return ret


@pytest.mark.parametrize("testdata", _get_isodatapaths(), ids=lambda d: d[0])
def test_iso_detection(testdata):
    osname, isodatapaths = testdata
    for isodatapath in isodatapaths:
        # A symlink indicates an ISO from another version
        # that is ambiguous and so accidentally matches
        # this distro too. We ignore them here, because
        # we'll validate against the primary OS version
        if isodatapath.is_symlink():
            continue

        detected = []
        isodatamedia = isodata.get_isodatamedia(isodatapath)
        for osxml2 in util.DataFiles.oses():
            for media in osxml2.medias:
                if isodatamedia.match(media.iso):
                    if osname != osxml2.shortid:
                        logging.warning(
                            "ISO '%s' was matched by OS '%s' while it "
                            "should only be matched by OS '%s'",
                            isodatamedia.path,
                            osxml2.shortid,
                            osname,
                        )
                    else:
                        logging.info(
                            "ISO '%s' matched by OS '%s'",
                            isodatamedia.path,
                            osxml2.shortid,
                        )
                    # For several distros we do not have the volume-size
                    # set as part of our DB, thus multiple detections may
                    # occur. Although this case is not the optimal, as long
                    # as we detect the very same distro it's okay-ish.
                    if osxml2.shortid not in detected:
                        detected.append(osxml2.shortid)

        ok = True
        for osid in detected:
            # Primary (correct) match
            if osid == osname:
                continue

            # Ambiguous (incorrect) matches, where we can't
            # distinguish from a different distro version
            link = Path(isodatapath.parent.parent, osid, isodatapath.name)
            if not link.exists():
                ok = False
        if ok:
            continue

        raise AssertionError(
            "isodata: %s\nMatched=%s but expected=%s"
            % (isodatapath, detected, [osname])
        )



#!/usr/bin/env python3
"""
Refreshes data from the index of downloadable images published by Endless.

Each eos3.X branch gets a separate entry in the database; when a new release is made in
that series, the URLs are refreshed to mark the newest point release. When a new branch
is released, it is annotated as updating and deriving from the previous branch, and the
old branch is marked as end-of-life as of the first release on the new branch.
"""
# Copyright © 2019 Endless Mobile, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import argparse
import datetime
from distutils.version import LooseVersion
import json
import jinja2
import os
import requests
import subprocess
import tempfile

import gi

gi.require_version("GnomeDesktop", "3.0")
from gi.repository import GnomeDesktop  # noqa: E402

BASE_URL = "https://images-dl.endlessm.com"
MANIFEST_URL = BASE_URL + "/releases-eos-3.json"
DATA_DIR = os.path.relpath(
    os.path.abspath(
        os.path.join(
            os.path.dirname(__file__), "..", "..", "data", "os", "endlessos.com"
        )
    )
)
ISODATA_DIR = os.path.relpath(
    os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "..", "tests", "isodata", "eos")
    )
)

# Endless OS images have a "personality", which determines which content they have
# pre-installed and perhaps some settings overrides. (The base OS is the same in all
# cases.) For our downloadable images, the personality is currently either a locale ID
# ('en', 'pt_BR') or 'base' (which contains almost no pre-installed apps).
NON_LOCALE_PERSONALITIES = {"base": "Basic", "sea": "Southeast Asia"}


def personality_name(personality):
    # TODO: this is adapted from our installer, but we should include this information
    # in the releases-eos-3.json file.
    try:
        return NON_LOCALE_PERSONALITIES[personality]
    except KeyError:
        pass

    # Get the name of 'personality' in US English. This data actually comes from the
    # iso-codes package, in the iso_639 and iso_3166 gettext domains, with extra logic
    # to suppress the country name if the language is "unique".
    #
    # Without appending .UTF-8, these come out as (eg):
    #   Portuguese (Brazil) [ISO-8859-1]
    personality_as_locale = f"{personality}.UTF-8"
    locale_name = GnomeDesktop.get_language_from_locale(personality_as_locale, "en_US")
    if locale_name:
        # TODO: consider using the same API to generate translations from
        # "Endless OS Portuguese (Brazil)" to "Endless OS Portugais (Brésil)". At
        # present, we do not transliterate "Endless OS" to (eg) "Ендлесс ОС" but if this
        # changed we would need to be a bit careful.
        return locale_name

    return personality


def publisher_id(branch):
    if LooseVersion(branch) >= LooseVersion("3.9"):
        return "ENDLESS OS FOUNDATION LLC"
    else:
        return "ENDLESS COMPUTERS"


def series(branch):
    eos_prefix_len = len("eos")
    return branch[eos_prefix_len:]


def adj(series, delta):
    major, minor = series.split(".")
    previous_minor = int(minor) + delta
    return f"{major}.{previous_minor}"


def predecessor(series):
    # Branching/Numbering scheme changed slightly in the transition from
    # eos3.9 → eos4.0 and eos5.0
    if series == "4.0":
        return "3.9"
    if series == "5.0":
        return "4.0"
    return adj(series, -1)


def successor(series):
    # Branching/Numbering scheme changed slightly in the transition from
    # eos3.9 → eos4.0 and eos5.0
    if series == "3.9":
        return "4.0"
    if series == "4.0":
        return "5.0"

    return adj(series, 1)


def guess_release_date(image):
    release_dates = [
        datetime.datetime.fromisoformat(i["iso"]["last_modified"]).date()
        for i in image["personality_images"].values()
        if "iso" in i
    ]
    return min(release_dates, default=None)


def all_personalities_for_branch(images):
    # We sometimes add and remove personalities within a series. For example, in
    # 3.4.5 we removed ar, bn, es_GT, es_MX and zh_CN from the set of downloadable
    # images.
    return {
        personality for image in images for personality in image["personality_images"]
    }


def fetch_isodata(branch, iso):
    iso_url = BASE_URL + "/" + iso["file"]
    isodata_file = os.path.join(ISODATA_DIR, branch, os.path.basename(iso_url) + ".txt")

    if os.path.exists(isodata_file):
        return
    print(f"Fetching first 2MiB of {iso_url}...")

    response = requests.get(iso_url, headers={"Range": "bytes=0-2097152"}, stream=True)
    with tempfile.NamedTemporaryFile(suffix=os.path.basename(iso_url)) as f:
        for chunk in response.iter_content(chunk_size=8096):
            f.write(chunk)

        f.flush()

        os.makedirs(os.path.dirname(isodata_file), exist_ok=True)
        with open(isodata_file, "w") as g:
            subprocess.run(
                ("isoinfo", "-d", "-i", f.name),
                stdout=g,
                check=True,
            )


def fetch_all_isodata(image):
    for personality_images in image["personality_images"].values():
        fetch_isodata(image["branch"], personality_images["iso"])


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--releases-json",
        type=argparse.FileType("r"),
        help="path to releases-eos-3.json file (default: fetch from network)",
    )
    args = parser.parse_args()

    if args.releases_json is not None:
        manifest = json.load(args.releases_json)
    else:
        response = requests.get(MANIFEST_URL)
        response.raise_for_status()
        manifest = response.json()

    images = list(manifest["images"].values())
    images.sort(key=lambda i: LooseVersion(i["version"]))
    images_by_branch = {}
    for image in images:
        images_by_branch.setdefault(image["branch"], []).append(image)

    # Prepare template
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(DATA_DIR),
        autoescape=True,
        undefined=jinja2.StrictUndefined,
    )
    env.filters["personality_name"] = personality_name
    env.filters["publisher_id"] = publisher_id
    template = env.get_template("eos-3.xml.in.in")

    for branch in images_by_branch:
        oldest_image = images_by_branch[branch][0]
        image = images_by_branch[branch][-1]
        release_date = guess_release_date(oldest_image)
        if not release_date:
            # eos3.0 did not have ISOs; pretend it never existed!
            continue

        current_series = series(image["branch"])

        previous_series = predecessor(current_series)
        if guess_release_date(images_by_branch[f"eos{previous_series}"][-1]) is None:
            # eos3.0 did not have ISOs; pretend it never existed!
            previous_series = None

        next_series = successor(current_series)
        try:
            next_series_images = images_by_branch[f"eos{next_series}"]
        except KeyError:
            eol_date = None
        else:
            # Date of first release in next series
            eol_date = guess_release_date(next_series_images[0])

        retired_personalities = (
            all_personalities_for_branch(images_by_branch[branch])
            - image["personality_images"].keys()
        )

        xml = template.render(
            base_url=BASE_URL,
            image=image,
            release_date=release_date,
            current_series=current_series,
            previous_series=previous_series,
            eol_date=eol_date,
            retired_personalities=retired_personalities,
        )
        with open(os.path.join(DATA_DIR, f"eos-{current_series}.xml.in"), "w") as f:
            f.write(xml)

        if eol_date is None:
            fetch_all_isodata(image)

        # We had an entry for eos3.10 in osinfo-db, which actually ended up being
        # released as eos4.0. Since we can never remove entries from osinfo-db,
        # generate it as an EOLed clone of eos4.0, with no images.
        if current_series == "4.0":
            xml = template.render(
                base_url=BASE_URL,
                image={"branch": "eos3.10", "personality_images": {}},
                release_date=None,
                current_series="3.10",
                previous_series=previous_series,
                eol_date=release_date,
                retired_personalities=[],
            )
            with open(os.path.join(DATA_DIR, "eos-3.10.xml.in"), "w") as f:
                f.write(xml)


if __name__ == "__main__":
    main()



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import re

from . import util


@util.os_parametrize("osxml")
def test_validate_short_ids(osxml):
    invalids = []

    pattern = r"[a-z0-9\.\-]"
    for shortid in osxml.shortids:
        result = re.match(pattern, shortid)
        if not result:
            invalids.append(shortid)

    assert invalids == []
